from __future__ import annotations

__all__ = ["ParamUDCT", "udctmddec", "udctmdrec", "udctmdwin"]
from curvelets.reference.udct import ParamUDCT, udctmddec, udctmdrec
from curvelets.reference.udctmdwin import udctmdwin



from __future__ import annotations

from math import prod

import numpy as np


def downsamp(F, decim):
    assert F.ndim == len(decim)
    D = F[tuple(slice(None, None, d) for d in decim)].copy()
    return D


def downsamp_old_not_working(F, decim):
    # Downsampling F in decim
    # downsamp can work in any dimension
    # decim must be vector equal to dimension of F
    sz = F.shape
    # max value
    ms = prod(sz)
    indw = 1 + np.arange(ms)

    dim = len(sz)
    mask = np.zeros((ms, dim))

    for in1 in range(1, dim + 1):
        if in1 == 1:
            step = decim[in1 - 1]
            template = np.zeros((step, 1))
            template.flat[0] = 1
        else:
            step = decim[in1 - 1]
            tmp = np.zeros((step, 1))
            tmp.flat[0] = 1
            tmp2 = np.ones((prod(sz[: in1 - 1]), 1))
            template = np.kron(tmp, tmp2)

        kronsz = np.ones((ms // len(template), 1))
        maskcol = np.kron(kronsz, template)
        mask[:, in1 - 1 : in1] = maskcol

    maskp = np.prod(mask, axis=1)
    idx = np.nonzero(maskp.ravel())[0]
    indwf = indw.flat[idx]

    D = np.reshape(F.flat[indwf - 1], sz // decim)
    return D


def downsamp_save(F, decim):
    assert F.ndim == len(decim)
    # Downsampling F in decim
    D = F[tuple(slice(None, None, d) for d in decim)].copy()
    return D



from __future__ import annotations

import numpy as np

from curvelets.utils import array_split_nd, deepflatten, ndargmax


def test_ndargmax():
    x = np.zeros((10, 10, 10))
    x[1, 1, 1] = 1.0
    m = ndargmax(x)
    assert len(m) == 3
    assert np.unique(m).item() == 1


def test_deepflatten():
    flattened = list(deepflatten([[[[[[1], 2, [[3, 4]]]]]]]))
    np.testing.assert_equal(flattened, np.arange(1, 5))


def test_array_split_nd():
    ary = np.outer(1 + np.arange(2), 2 + np.arange(3))
    split = array_split_nd(ary, 2, 3)
    desired = [
        [np.array([[2]]), np.array([[3]]), np.array([[4]])],
        [np.array([[4]]), np.array([[6]]), np.array([[8]])],
    ]
    assert len(split) == len(desired)
    for s, d in zip(split, desired):
        assert len(s) == len(d)
    for si, di in zip(deepflatten(split), deepflatten(desired)):
        np.testing.assert_equal(si, di)


def test_apply_along_wedges():
    import numpy as np

    from curvelets.numpy import SimpleUDCT
    from curvelets.utils import apply_along_wedges

    x = np.zeros((32, 32))
    C = SimpleUDCT(x.shape, nscales=3, nbands_per_direction=3)
    y = C.forward(x)
    res = apply_along_wedges(y, lambda w, *_: w.shape)

    for si, di in zip(deepflatten(res), deepflatten(y)):
        np.testing.assert_equal(np.array(si), np.array(di.shape))



from __future__ import annotations

import sys
from dataclasses import dataclass, field
from math import ceil, prod

import numpy as np
import numpy.typing as npt


@dataclass(**({"kw_only": True} if sys.version_info >= (3, 10) else {}))
class ParamUDCT:
    dim: int
    size: tuple[int, ...]
    cfg: npt.NDArray[np.int_]  # last dimension  == dim
    alpha: float
    r: tuple[float, float, float, float]
    winthresh: float
    len: int = field(init=False)
    res: int = field(init=False)
    decim: npt.NDArray[np.int_] = field(init=False)
    ind: dict[int, dict[int, np.ndarray]] | None = None
    dec: dict[int, np.ndarray] | None = None

    def __post_init__(self) -> None:
        self.len = prod(self.size)
        self.res = len(self.cfg)
        self.decim = 2 * (np.asarray(self.cfg, dtype=int) // 3)


def circshift(arr: np.ndarray, shape: tuple[int, ...]) -> np.ndarray:
    assert arr.ndim == len(shape)
    return np.roll(arr, shape, axis=tuple(range(len(shape))))


def adapt_grid(S1: np.ndarray, S2: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    x1, x2 = np.meshgrid(S2, S1)

    t1: npt.NDArray[np.floating] = np.zeros_like(x1, dtype=float)
    ind = (x1 != 0) & (np.abs(x2) <= np.abs(x1))
    t1[ind] = -x2[ind] / x1[ind]

    t2: npt.NDArray[np.floating] = np.zeros_like(x1, dtype=float)
    ind = (x2 != 0) & (np.abs(x1) < np.abs(x2))
    t2[ind] = x1[ind] / x2[ind]

    t3 = t2.copy()
    t3[t2 < 0] = t2[t2 < 0] + 2
    t3[t2 > 0] = t2[t2 > 0] - 2

    M1 = t1 + t3
    M1[x1 >= 0] = -2

    t1 = np.zeros_like(x1, dtype=float)
    ind = (x2 != 0) & (abs(x1) <= abs(x2))
    t1[ind] = -x1[ind] / x2[ind]

    t2 = np.zeros_like(x1, dtype=float)
    ind = (x1 != 0) & (abs(x2) < abs(x1))
    t2[ind] = x2[ind] / x1[ind]

    t3 = t2.copy()
    t3[t2 < 0] = t2[t2 < 0] + 2
    t3[t2 > 0] = t2[t2 > 0] - 2

    M2 = t1 + t3
    M2[x2 >= 0] = -2

    return M2, M1


def angle_fun(Mgrid: np.ndarray, direction: int, n: int, alpha: float) -> np.ndarray:
    # % create 2-D grid function-------------------------------------------------

    # angle meyer window
    angd = 2 / n
    ang = angd * np.array([-alpha, alpha, 1 - alpha, 1 + alpha])

    Mang = []
    # This is weird, both directions are the same code
    if direction in (1, 2):
        for jn in range(1, ceil(n / 2) + 1):
            ang2 = -1 + (jn - 1) * angd + ang
            fang = fun_meyer(Mgrid, *ang2)
            Mang.append(fang[None, :])
    else:
        msg = "Unrecognized direction"
        raise ValueError(msg)
    return np.concatenate(Mang, axis=0)


def angle_kron(
    angle_arr: np.ndarray, nper: np.ndarray, param_udct: ParamUDCT
) -> np.ndarray:
    # , nper, param_udct
    krsz: npt.NDArray[np.int_] = np.ones(3, dtype=int)
    krsz[0] = np.prod(param_udct.size[: nper[0] - 1])
    krsz[1] = np.prod(param_udct.size[nper[0] : nper[1] - 1])
    krsz[2] = np.prod(param_udct.size[nper[1] : param_udct.dim])

    tmp1 = np.kron(np.ones((krsz[1], 1), dtype=int), angle_arr)
    tmp2 = np.kron(np.ones((krsz[2], 1), dtype=int), travel(tmp1)).ravel()
    tmp3 = travel(np.kron(tmp2, np.ones((krsz[0], 1), dtype=int)))
    return tmp3.reshape(*param_udct.size[::-1]).T


def downsamp(F: np.ndarray, decim: np.ndarray) -> np.ndarray:
    assert F.ndim == len(decim)
    return F[tuple(slice(None, None, d) for d in decim)]


def fftflip(F: np.ndarray, axis: int) -> np.ndarray:
    Fc = F
    dim = F.ndim
    shiftvec: npt.NDArray[np.int_] = np.zeros((dim,), dtype=int)
    shiftvec[axis] = 1
    Fc = np.flip(F, axis)
    return circshift(Fc, tuple(shiftvec))


def fun_meyer(x: np.ndarray, p1: float, p2: float, p3: float, p4: float) -> np.ndarray:
    p = np.array([-20.0, 70.0, -84.0, 35.0, 0.0, 0.0, 0.0, 0.0])
    y = np.zeros_like(x)

    win = (x >= p1) & (x <= p2)
    y[win] = np.polyval(p, (x[win] - p1) / (p2 - p1))

    win = (x > p2) & (x <= p3)
    y[win] = 1.0

    win = (x >= p3) & (x <= p4)
    y[win] = np.polyval(p, (x[win] - p4) / (p3 - p4))
    return y


def travel(arr: np.ndarray) -> np.ndarray:
    return arr.T.ravel()


def travel_new(arr: np.ndarray) -> np.ndarray:
    return arr.ravel()


def to_sparse(arr: np.ndarray, thresh: float) -> np.ndarray:
    idx = np.argwhere(travel(arr) > thresh)
    out: npt.NDArray[np.floating] = np.c_[idx + 1, travel(arr)[idx]]
    return out


def to_sparse_new(arr: np.ndarray, thresh: float) -> list[np.ndarray]:
    idx = np.argwhere(arr.ravel() > thresh)
    return [idx, arr.ravel()[idx]]


def from_sparse(arr: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    idx = arr[:, 0].astype(int) - 1
    val = arr[:, 1]
    return idx, val


def from_sparse_new(arr_list: list[np.ndarray]) -> list[np.ndarray]:
    return arr_list


def upsamp(F: np.ndarray, decim: np.ndarray) -> np.ndarray:
    assert F.ndim == len(decim)
    upsamp_shape = tuple(s * d for s, d in zip(F.shape, decim))
    D = np.zeros(upsamp_shape, dtype=F.dtype)
    D[tuple(slice(None, None, d) for d in decim)] = F[...]
    return D



from __future__ import annotations

from math import ceil

import numpy as np

from ._fun_meyer import fun_meyer


def angle_fun(Mgrid, direction, n, alpha):
    # % create 2-D grid function-------------------------------------------------

    # angle meyer window
    angd = 2 / n
    ang = angd * np.array([-alpha, alpha, 1 - alpha, 1 + alpha])

    Mang = []
    if direction == 1:
        for jn in range(1, ceil(n / 2) + 1):
            ang2 = -1 + (jn - 1) * angd + ang
            fang = fun_meyer(Mgrid, *ang2)
            Mang.append(fang[None, :])
    elif direction == 2:  # This is weird, both directions are the same code
        for jn in range(1, ceil(n / 2) + 1):
            ang2 = -1 + (jn - 1) * angd + ang
            fang = fun_meyer(Mgrid, *ang2)
            Mang.append(fang[None, :])
    else:
        raise ValueError("unrecognized direction")
    return np.concatenate(Mang, axis=0)



from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from matplotlib.cm import ScalarMappable
from matplotlib.colors import Normalize

from curvelets.utils import deepflatten


def overlay_disk(
    c_struct: list[list[list[float]]],
    ax: Axes | None = None,
    linewidth: float = 5,
    linecolor: str = "r",
    cmap: str = "turbo",
    vmin: float | None = None,
    vmax: float | None = None,
    direction: str = "normal",
) -> Axes:
    r"""Overlay a curvelet structure onto its locations on a multiscale, multidirectional disk.

    Its intended usage is to display various scalars derived from curvelet
    wedges of a certain image with a disk display.


    Parameters
    ----------
    c_struct : list[list[list[float]]]
        A scale for every curvelet wedge.
    ax : :obj:`Axes <matplotlib.axes.Axes>`, optional
        Axis on which to overlay the disk. Uses :obj:`plt.gca <matplotlib.pyplot.gca>` if None.
    linewidth : float, optional
        Width of line separating scales as a percentage of a single scale height, by default 5.
        Set to zero to disable.
    linecolor : str, optional
        Color of line separating scales, by default "r".
    cmap : str, optional
        Colormap, by default ``"turbo"``.
    vmin, vmax : float, optional
        Data range that the colormap covers. By default, the colormap covers the
        complete value range of the supplied data.
    direction : str, optional
        "tangent" or, by default "normal"

    Returns
    -------
    :obj:`Axes <matplotlib.axes.Axes>`
        Axis used.
    """
    ax = plt.gca() if ax is None else ax
    ax.axis("off")

    if vmin is None:
        vmin = min(v for v in deepflatten(c_struct))
    if vmax is None:
        vmax = max(v for v in deepflatten(c_struct))
    cmapper = ScalarMappable(norm=Normalize(vmin, vmax, clip=True), cmap=cmap)

    nscales = len(c_struct)
    ndir = 2  # Only available for 2D!

    deg_360 = 2 * np.pi
    deg_135 = np.pi * 3 / 4
    deg_n45 = -np.pi / 4
    deg_90 = np.pi / 2

    linewidth *= 0.01 / (nscales - 1)
    wedge_height = 1 / (nscales - 1)
    magic_shift = 0  # 0 or -np.pi/8 or -np.pi/16? something else?
    for iscale in range(nscales):
        ndir = len(c_struct[iscale])
        if iscale > 0:
            assert ndir == 2
        for idir in range(ndir):
            nwedges = len(c_struct[iscale][idir])
            angles_per_wedge = deg_90 / nwedges
            pm = (-1) ** (idir + 1)  # CC for idir == 0, CCW otherwise
            for iwedge in range(nwedges):
                color = cmapper.to_rgba(c_struct[iscale][idir][iwedge])
                if iscale == 0:
                    ax.bar(
                        x=0, height=wedge_height, width=deg_360, bottom=0, color=color
                    )
                    continue
                for offset in [deg_135, deg_n45]:  # top-left, bottom-right
                    wedge_x = (
                        offset + pm * angles_per_wedge * (0.5 + iwedge) + magic_shift
                    )
                    if direction == "tangent":
                        wedge_x += deg_90
                    wedge_width = angles_per_wedge
                    wedge_bottom = iscale * wedge_height
                    ax.bar(
                        x=wedge_x,
                        height=wedge_height,
                        width=wedge_width,
                        bottom=wedge_bottom,
                        color=color,
                    )
    if linewidth == 0:
        return ax
    # Plot after so they are on top
    for iscale in range(nscales):
        # Scale separators
        ax.bar(
            x=0,
            height=linewidth,
            width=deg_360,
            bottom=(iscale + 1 - linewidth / 2) / (nscales - 1),
            color=linecolor,
        )
        if iscale == 0:
            continue
        # Wedge separators
        for idir in range(len(c_struct[iscale])):
            nwedges = len(c_struct[iscale][idir])
            angles_per_wedge = deg_90 / nwedges
            pm = (-1) ** (idir + 1)
            for iwedge in range(nwedges):
                for offset in [deg_135, deg_n45]:  # top-left, bottom-right
                    wedge_x = (
                        offset + pm * angles_per_wedge * (0.5 + iwedge) + magic_shift
                    )
                    if direction == "tangent":
                        wedge_x += deg_90
                    wedge_width = angles_per_wedge
                    wedge_bottom = iscale * wedge_height
                    ax.bar(
                        x=wedge_x - wedge_width / 2,
                        height=wedge_height,
                        width=linewidth,
                        bottom=wedge_bottom,
                        color=linecolor,
                    )

    return ax



# import matplotlib.pyplot as plt
from __future__ import annotations

import numpy as np

from ._travel import travel


def angle_kron(angle_fun, nper, param_udct):
    # , nper, param_udct
    krsz = np.ones(3, dtype=int)
    krsz[0] = np.prod(param_udct.size[: nper[0] - 1])
    krsz[1] = np.prod(param_udct.size[nper[0] : nper[1] - 1])
    krsz[2] = np.prod(param_udct.size[nper[1] : param_udct.dim])

    tmp = angle_fun

    tmp1 = np.kron(np.ones((krsz[1], 1), dtype=int), tmp)
    tmp2 = np.kron(np.ones((krsz[2], 1), dtype=int), travel(tmp1)).ravel()
    tmp3 = travel(np.kron(tmp2, np.ones((krsz[0], 1), dtype=int)))
    ang_md = tmp3.reshape(*param_udct.size[::-1]).T

    return ang_md



from __future__ import annotations

from typing import Any

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from matplotlib.colorbar import Colorbar
from matplotlib.image import AxesImage
from mpl_toolkits.axes_grid1 import make_axes_locatable
from numpy.typing import NDArray


def despine(ax: Axes) -> None:
    """Remove the top and right spines from plot(s).

    Parameters
    ----------
    ax : :obj:`Axes <matplotlib.axes.Axes>`
        Axis to despine.
    """
    for spine in ax.spines:
        ax.spines[spine].set_visible(False)
    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)


def create_colorbar(
    im: AxesImage,
    ax: Axes | None = None,
    size: float = 0.05,
    pad: float = 0.1,
    orientation: str = "vertical",
) -> tuple[Axes, Colorbar]:
    r"""Create a colorbar. Divides axis and attaches a colorbar to it.

    Parameters
    ----------
    im : :obj:`AxesImage <matplotlib.image.AxesImage>`
        Image from which the colorbar will be created.
        Commonly the output of :obj:`plt.imshow <matplotlib.pyplot.imshow>`.
    ax : :obj:`Axes <matplotlib.axes.Axes>`, optional
        Axis which to split. Uses :obj:`plt.gca <matplotlib.pyplot.gca>` if None.
    size : :obj:`float`, optional
        Size of split, by default 0.05. Effectively sets the size of the colorbar.
    pad : :obj:`float`, optional
        Padding between colorbar axis and input axis, by default 0.1.
    orientation : :obj:`str`, optional
        Orientation of the colorbar, by default "vertical".

    Returns
    -------
    cax : :obj:`Axes <matplotlib.axes.Axes>`
        Colorbar axis.
    cb : :obj:`Colorbar <matplotlib.colorbar.Colorbar>`
        Colorbar.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> from matplotlib.ticker import MultipleLocator
    >>> from curvelets.plot import create_colorbar
    >>> fig, ax = plt.subplots()
    >>> im = ax.imshow([[0]], vmin=-1, vmax=1, cmap="gray")
    >>> cax, cb = create_colorbar(im, ax)
    >>> cax.yaxis.set_major_locator(MultipleLocator(0.1))
    >>> print(cb.vmin)
    -1.0
    """
    ax = plt.gca() if ax is None else ax
    divider = make_axes_locatable(ax)
    cax: Axes = divider.append_axes("right", size=f"{size:%}", pad=pad)
    fig = ax.get_figure()
    if fig is None:
        cb = plt.colorbar(im, cax=cax, orientation=orientation)
    else:
        cb = fig.colorbar(im, cax=cax, orientation=orientation)
    return cax, cb


def _create_range(start: float, end: float, n: int) -> NDArray[np.float64]:
    return start + (end - start) * (0.5 + np.arange(n)) / n


def create_inset_axes_grid(
    ax: Axes,
    rows: int = 1,
    cols: int = 1,
    height: float = 0.5,
    width: float = 0.5,
    squeeze: bool = True,
    kwargs_inset_axes: dict[str, Any] | None = None,
) -> NDArray[np.object_] | Axes:
    r"""Create a grid of insets.

    The input axis will be overlaid with a grid of insets.
    Numbering of the axes is top to bottom (rows) and
    left to right (cols).

    Parameters
    ----------
    ax : :obj:`Axes <matplotlib.axes.Axes>`
        Input axis.
    rows : :obj:`int`, optional
        Number of rows, by default 1.
    cols : :obj:`int`, optional
        Number of columns, by default 1.
    width : :obj:`float`, optional
        Width of each axis, as a percentage of ``cols``, by default 0.5.
    height : :obj:`float`, optional
        Height of each axis, as a percentage of ``rows``, by default 0.5.
    squeeze : :obj:`float`
        If True, removes singleton dimensions like :obj:`plt.subplots <matplotlib.pyplot.subplots>`.
    kwargs_inset_axes : :obj:`dict[str, Any]`, optional
        Arguments to be passed to :obj:`matplotlib.axes.Axes.inset_axes`.

    Returns
    -------
    :obj:`NDArray <numpy.typing.NDArray>` [:obj:`Axes <matplotlib.axes.Axes>`]
        Array of :obj:`Axes <matplotlib.axes.Axes>` shaped ``(rows, cols)``.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> import numpy as np
    >>> from curvelets.plot import create_inset_axes_grid
    >>> fig, ax = plt.subplots(figsize=(6, 6))
    >>> ax.imshow([[0]], extent=[-2, 2, 2, -2], vmin=-1, vmax=1, cmap="gray")
    >>> rows, cols = 2, 3
    >>> inset_axes = create_inset_axes_grid(
    >>>     ax,
    >>>     rows,
    >>>     cols,
    >>>     width=0.5,
    >>>     height=0.5,
    >>>     kwargs_inset_axes={"projection": "polar"},
    >>> )
    >>> nscales = 4
    >>> lw = 0.1
    >>> for irow in range(rows):
    >>>     for icol in range(cols):
    >>>         for iscale in range(1, nscales):
    >>>             inset_axes[irow][icol].bar(
    >>>                 x=0,
    >>>                 height=lw,
    >>>                 width=2 * np.pi,
    >>>                 bottom=((iscale + 1) - 0.5 * lw) / (nscales - 1),
    >>>                 color="r",
    >>>             )
    >>>             inset_axes[irow][icol].set(title=f"Row, Col: ({irow}, {icol})")
    >>>             inset_axes[irow][icol].axis("off")
    """
    if kwargs_inset_axes is None:
        kwargs_inset_axes = {}

    axes: NDArray[np.object_] = np.empty((rows, cols), dtype=Axes)

    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    xmin, xmax = min(xmin, xmax), max(xmin, xmax)
    ymin, ymax = min(ymin, ymax), max(ymin, ymax)

    width *= (xmax - xmin) / cols
    height *= (ymax - ymin) / rows

    for irow, rowpos in enumerate(_create_range(ymin, ymax, rows)):
        for icol, colpos in enumerate(_create_range(xmin, xmax, cols)):
            axes[irow, icol] = ax.inset_axes(
                (colpos - 0.5 * width, rowpos - 0.5 * height, width, height),
                transform=ax.transData,
                **kwargs_inset_axes,
            )
    if squeeze:
        if rows == cols == 1:
            return axes[0, 0]
        return axes.squeeze()
    return axes


def overlay_arrows(
    vectors: NDArray[np.generic], ax: Axes, arrowprops: dict[str, Any] | None = None
) -> Axes:
    r"""Overlay arrows on an axis.

    Parameters
    ----------
    vectors : :obj:`NDArray <numpy.typing.NDArray>`
        Array shaped ``(rows, cols, 2)``, corresponding to a 2D vector field.
    ax : :obj:`Axes <matplotlib.axes.Axes>`
        Axis on which to overlay the arrows.
    arrowprops : :obj:`dict`, optional
        Arrow properties, to be passed to :obj:`matplotlib.pyplot.annotate`.
        By default will be set to ``{"facecolor": "black", "shrink": 0.05}``.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> import numpy as np
    >>> from curvelets.plot import overlay_arrows
    >>> fig, ax = plt.subplots(figsize=(8, 10))
    >>> ax.imshow([[0]], vmin=-1, vmax=1, extent=[0, 1, 1, 0], cmap="gray")
    >>> rows, cols = 3, 4
    >>> kvecs = np.array(
    >>>     [
    >>>         [(1 + x, x * y) for x in (0.5 + np.arange(cols)) / cols]
    >>>         for y in (0.5 + np.arange(rows)) / rows
    >>>     ]
    >>> )
    >>> overlay_arrows(
    >>>     0.05 * kvecs,
    >>>     ax,
    >>>     arrowprops=dict(
    >>>         facecolor="r",
    >>>         shrink=0.05,
    >>>         width=10 / cols,
    >>>         headwidth=10,
    >>>         headlength=10,
    >>>     ),
    >>> )
    """
    rows, cols, _ = vectors.shape

    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    xmin, xmax = min(xmin, xmax), max(xmin, xmax)
    ymin, ymax = min(ymin, ymax), max(ymin, ymax)

    if arrowprops is None:
        arrowprops = {"facecolor": "black", "shrink": 0.05}

    for irow, rowpos in enumerate(_create_range(ymin, ymax, rows)):
        for icol, colpos in enumerate(_create_range(xmin, xmax, cols)):
            ax.annotate(
                "",
                xy=(
                    colpos + vectors[irow, icol, 0],
                    rowpos + vectors[irow, icol, 1],
                ),
                xytext=(colpos, rowpos),
                xycoords="data",
                arrowprops=arrowprops,
                annotation_clip=False,
            )
    return ax



from __future__ import annotations

import sys

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias


import numpy as np
import numpy.typing as npt

if sys.version_info <= (3, 9):
    from typing import List  # noqa: UP035

    UDCTCoefficients: TypeAlias = List[List[List[npt.NDArray[np.complexfloating]]]]  # noqa: UP006
    UDCTWindows: TypeAlias = List[List[List[List[npt.NDArray[np.floating]]]]]  # noqa: UP006
else:
    UDCTCoefficients: TypeAlias = list[list[list[npt.NDArray[np.complexfloating]]]]
    UDCTWindows: TypeAlias = list[list[list[list[npt.NDArray[np.floating]]]]]



"""
Copyright (c) 2024 Carlos Alberto da Costa Filho. All rights reserved.

Curvelets: Python package for curvelet transforms.
"""

from __future__ import annotations

from ._version import version as __version__

__all__ = ["__version__", "numpy", "reference"]

from . import numpy, reference



from __future__ import annotations

__all__ = [
    "apply_along_wedges",
    "array_split_nd",
    "deepflatten",
    "ndargmax",
    "normal_vector_field",
]
from ._utils import (
    apply_along_wedges,
    array_split_nd,
    deepflatten,
    ndargmax,
    normal_vector_field,
)



from __future__ import annotations

import logging
from copy import deepcopy
from math import prod

import numpy as np
import numpy.typing as npt

from .typing import UDCTCoefficients, UDCTWindows
from .udctmdwin import udctmdwin
from .utils import ParamUDCT, downsamp, from_sparse_new, upsamp


def udctmddec(
    im: np.ndarray,
    param_udct: ParamUDCT,
    udctwin: UDCTWindows,
    decimation_ratio: list[npt.NDArray[np.int_]],
) -> UDCTCoefficients:
    imf = np.fft.fftn(im)
    cdtype = imf.dtype

    fband = np.zeros_like(imf)
    idx, val = from_sparse_new(udctwin[0][0][0])
    fband.flat[idx] = imf.flat[idx] * val.astype(cdtype)
    cband = np.fft.ifftn(fband)

    coeff: UDCTCoefficients = [[[downsamp(cband, decimation_ratio[0][0])]]]
    norm = np.sqrt(
        np.prod(np.full((param_udct.dim,), fill_value=2 ** (param_udct.res - 1)))
    )
    coeff[0][0][0] *= norm

    for ires in range(1, 1 + param_udct.res):
        coeff.append([])
        for idir in range(param_udct.dim):
            coeff[ires].append([])
            for iang in range(len(udctwin[ires][idir])):
                fband = np.zeros_like(imf)
                idx, val = from_sparse_new(udctwin[ires][idir][iang])
                fband.flat[idx] = imf.flat[idx] * val.astype(cdtype)

                cband = np.fft.ifftn(fband)
                decim = decimation_ratio[ires][idir, :]
                coeff[ires][idir].append(downsamp(cband, decim))
                coeff[ires][idir][iang] *= np.sqrt(2 * np.prod(decim))
    return coeff


def udctmdrec(
    coeff: UDCTCoefficients,
    param_udct: ParamUDCT,
    udctwin: UDCTWindows,
    decimation_ratio: list[npt.NDArray[np.int_]],
) -> np.ndarray:
    rdtype = coeff[0][0][0].real.dtype
    cdtype = (np.ones(1, dtype=rdtype) + 1j * np.ones(1, dtype=rdtype)).dtype
    imf = np.zeros(param_udct.size, dtype=cdtype)

    for ires in range(1, 1 + param_udct.res):
        for idir in range(param_udct.dim):
            for iang in range(len(udctwin[ires][idir])):
                decim = decimation_ratio[ires][idir, :]
                cband = upsamp(coeff[ires][idir][iang], decim)
                cband /= np.sqrt(2 * np.prod(decim))
                cband = np.prod(decim) * np.fft.fftn(cband)
                idx, val = from_sparse_new(udctwin[ires][idir][iang])
                imf.flat[idx] += cband.flat[idx] * val.astype(cdtype)

    imfl = np.zeros(param_udct.size, dtype=cdtype)
    decim = decimation_ratio[0][0]
    cband = upsamp(coeff[0][0][0], decim)
    cband = np.sqrt(np.prod(decim)) * np.fft.fftn(cband)
    idx, val = from_sparse_new(udctwin[0][0][0])
    imfl.flat[idx] += cband.flat[idx] * val.astype(cdtype)
    imf = 2 * imf + imfl
    return np.fft.ifftn(imf).real


class UDCT:
    def __init__(
        self,
        shape: tuple[int, ...],
        cfg: np.ndarray | None = None,
        alpha: float = 0.15,
        r: tuple[float, float, float, float] | None = None,
        winthresh: float = 1e-5,
        transpose: bool = False,
    ) -> None:
        self.transpose = transpose
        self.shape = shape[::-1] if self.transpose else shape
        dim = len(self.shape)
        cfg1 = np.c_[np.ones((dim,)) * 3, np.ones((dim,)) * 6].T if cfg is None else cfg
        r1: tuple[float, float, float, float] = (
            tuple(np.array([1.0, 2.0, 2.0, 4.0]) * np.pi / 3) if r is None else r
        )
        self.params = ParamUDCT(
            dim=dim, size=self.shape, cfg=cfg1, alpha=alpha, r=r1, winthresh=winthresh
        )

        self.windows, self.decimation, self.indices = udctmdwin(self.params)

    def vect(self, coeffs: UDCTCoefficients) -> npt.NDArray[np.complexfloating]:
        coeffs_vec = []
        for c in coeffs:
            for d in c:
                for w in d:
                    if self.transpose:
                        w = np.transpose(w)  # noqa: PLW2901
                    coeffs_vec.append(w.ravel())
        return np.concatenate(coeffs_vec)

    def struct(self, coeffs_vec: npt.NDArray[np.complexfloating]) -> UDCTCoefficients:
        ibeg = 0
        coeffs: UDCTCoefficients = []
        for ires, decres in enumerate(self.decimation):
            coeffs.append([])
            for idir, decdir in enumerate(decres):
                coeffs[ires].append([])
                for _ in self.windows[ires][idir]:
                    shape_decim = self.shape // decdir
                    iend = ibeg + prod(shape_decim)
                    wedge = coeffs_vec[ibeg:iend].reshape(shape_decim)
                    if self.transpose:
                        wedge = np.transpose(wedge)
                    coeffs[ires][idir].append(wedge)
                    ibeg = iend
        return coeffs

    def forward(self, x: np.ndarray) -> UDCTCoefficients:
        if self.transpose:
            x = np.transpose(x)
        np.testing.assert_equal(self.shape, x.shape)
        c = udctmddec(x, self.params, self.windows, self.decimation)
        if self.transpose:
            for iscale, s in enumerate(c):
                for idir, d in enumerate(s):
                    for iwedge, w in enumerate(d):
                        c[iscale][idir][iwedge] = np.transpose(w)
        return c

    def backward(self, c: UDCTCoefficients) -> np.ndarray:
        if self.transpose:
            c = deepcopy(c)
            for iscale, s in enumerate(c):
                for idir, d in enumerate(s):
                    for iwedge, w in enumerate(d):
                        c[iscale][idir][iwedge] = np.transpose(w)
        x = udctmdrec(c, self.params, self.windows, self.decimation)
        if self.transpose:
            x = np.transpose(x)
        return x


class SimpleUDCT(UDCT):
    def __init__(
        self,
        shape: tuple[int, ...],
        nscales: int = 3,
        nbands_per_direction: int = 3,
        alpha: float | None = None,
        winthresh: float = 1e-5,
        transpose: bool = False,
    ) -> None:
        assert nscales > 1
        assert nbands_per_direction >= 3

        dim = len(shape)
        nbands: npt.NDArray[np.int_] = (
            nbands_per_direction * 2 ** np.arange(nscales - 1)
        ).astype(int)

        if alpha is None:
            if nbands_per_direction == 3:
                alpha = 0.15
            elif nbands_per_direction == 4:
                alpha = 0.3
            elif nbands_per_direction == 5:
                alpha = 0.5
            else:
                alpha = 0.5
        for i, nb in enumerate(nbands, start=1):
            if (const := 2 ** (i / nb) * (1 + 2 * alpha) * (1 + alpha)) >= nb:
                msg = (
                    f"alpha={alpha:.3f} does not respect the relationship "
                    f"(2^{i}/{nb})(1+2α)(1+α) = {const:.3f} = < 1 for scale {i+1}"  # noqa: RUF001
                )
                logging.warning(msg)
        cfg = np.tile(nbands[:, None], dim)
        r: tuple[float, float, float, float] = tuple(
            np.array([1.0, 2.0, 2.0, 4.0]) * np.pi / 3
        )

        super().__init__(
            shape=shape,
            cfg=cfg,
            alpha=alpha,
            r=r,
            winthresh=winthresh,
            transpose=transpose,
        )



from __future__ import annotations

import numpy as np

# from numba import njit
# from numpy.polynomial import Polynomial

# @njit
# def eval_meyer_polynomial(x):
#     return 35.0 * x**4 - 84.0 * x**5 + 70.0 * x**6 - 20.0 * x**7


# @njit
# def fun_meyer_numba(x, p1, p2, p3, p4):
#     y = np.zeros_like(x)

#     win = (x >= p1) & (x <= p2)
#     y[win] = eval_meyer_polynomial((x[win] - p1) / (p2 - p1))

#     win = (x > p2) & (x <= p3)
#     y[win] = 1.0

#     win = (x >= p3) & (x <= p4)
#     y[win] = eval_meyer_polynomial((x[win] - p4) / (p3 - p4))
#     return y


# def fun_meyer_new(x, p1, p2, p3, p4):
#     p = Polynomial([0.0, 0.0, 0.0, 0.0, 35.0, -84.0, 70.0, -20.0])
#     y = np.zeros_like(x)

#     win = (x >= p1) & (x <= p2)
#     y[win] = p((x[win] - p1) / (p2 - p1))

#     win = (x > p2) & (x <= p3)
#     y[win] = 1.0

#     win = (x >= p3) & (x <= p4)
#     y[win] = p((x[win] - p4) / (p3 - p4))
#     return y


def fun_meyer_original(x, p1, p2, p3, p4):
    p = np.array([-20.0, 70.0, -84.0, 35.0, 0.0, 0.0, 0.0, 0.0])
    y = np.zeros_like(x)

    win = (x >= p1) & (x <= p2)
    y[win] = np.polyval(p, (x[win] - p1) / (p2 - p1))

    win = (x > p2) & (x <= p3)
    y[win] = 1.0

    win = (x >= p3) & (x <= p4)
    y[win] = np.polyval(p, (x[win] - p4) / (p3 - p4))
    return y


fun_meyer = fun_meyer_original



from __future__ import annotations

__all__ = ["ParamUDCT", "udctmddec", "udctmdrec", "udctmdwin", "UDCT", "SimpleUDCT"]

from curvelets.numpy.udct import UDCT, SimpleUDCT, udctmddec, udctmdrec
from curvelets.numpy.udctmdwin import udctmdwin
from curvelets.numpy.utils import ParamUDCT



"""
Zone Plate
==========
"""

from __future__ import annotations

# %%
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import ticker

from curvelets.numpy.udct import UDCT
from curvelets.plot import create_colorbar, despine


def make_r(
    shape: tuple[int, ...], exponent: float = 1, origin: tuple[int, ...] | None = None
):
    orig = (
        tuple((np.asarray(shape).astype(float) - 1) / 2) if origin is None else origin
    )

    ramps = np.meshgrid(
        *[np.arange(s, dtype=float) - o for s, o in zip(shape, orig)], indexing="ij"
    )
    return sum(x**2 for x in ramps) ** (exponent / 2)


def make_zone_plate(shape: tuple[int, ...], amplitude: float = 1.0, phase: float = 0.0):
    mxsz = max(*shape)

    return amplitude * np.cos((np.pi / mxsz) * make_r(shape, 2) + phase)


# %%
# Setup
# #####

shape = (256, 256)
zone_plate = make_zone_plate(shape)
cfg = np.array([[3, 3], [6, 6], [12, 6]])
C = UDCT(shape=shape, cfg=cfg)

# %%
# Uniform Discrete Curvelet Transform Round Trip
# ##############################################

coeffs = C.forward(zone_plate)
zone_plate_inv = C.backward(coeffs)

# %%
vmax = np.abs(zone_plate).max()
opts = {"aspect": "equal", "cmap": "gray", "vmin": -vmax, "vmax": vmax}

fig, ax = plt.subplots(figsize=(4, 3))
im = ax.imshow(zone_plate.T, **opts)
_, cb = create_colorbar(im=im, ax=ax)
fmt = ticker.FuncFormatter(lambda x, _: f"{x:.0e}")
cb.ax.yaxis.set_major_formatter(fmt)
despine(ax)
ax.set(title="Input")

# %%
fig, ax = plt.subplots(figsize=(4, 3))
im = ax.imshow(zone_plate_inv.T, **opts)
_, cb = create_colorbar(im=im, ax=ax)
cb.ax.yaxis.set_major_formatter(fmt)
despine(ax)
ax.set(title="UDCT Round Trip")

# %%
opts["vmax"] = np.abs(zone_plate - zone_plate_inv).max()
opts["vmin"] = -opts["vmax"]
fig, ax = plt.subplots(figsize=(4, 3))
im = ax.imshow((zone_plate - zone_plate_inv).T, **opts)
_, cb = create_colorbar(im=im, ax=ax)
cb.ax.yaxis.set_major_formatter(fmt)
despine(ax)
ax.set(title=f"Error (max = {opts['vmax']:.2g})")

print(f"Max Error: {opts['vmax']:.2g}")  # noqa: T201

# %%
# Curvelet Coefficients: Amplitude and Phase
# ##########################################
for ires, _ in enumerate(coeffs):
    for idir, _ in enumerate(coeffs[ires]):
        for iang, _ in enumerate(coeffs[ires][idir]):
            z = coeffs[ires][idir][iang]
            opts["vmax"] = np.abs(z).max()
            opts["vmin"] = 0
            opts["cmap"] = "gray"
            fig, axs = plt.subplots(1, 2, figsize=(8, 3))
            im = axs[0].imshow(np.abs(z).T, **opts)
            _, cb = create_colorbar(im=im, ax=axs[0])
            fmt = ticker.FuncFormatter(lambda x, _: f"{x:.0e}")
            cb.ax.yaxis.set_major_formatter(fmt)
            opts["vmax"] = 180
            opts["vmin"] = -opts["vmax"]
            opts["cmap"] = "hsv"
            im = axs[1].imshow(np.angle(z, deg=True).T, **opts)
            _, cb = create_colorbar(im=im, ax=axs[1])
            cb.ax.yaxis.set_major_locator(ticker.MultipleLocator(45))
            cb.ax.yaxis.set_major_formatter(lambda x, _: f"{x:.0f}°")
            axs[0].set(title="Amplitude")
            axs[1].set(title="Phase")
            fig.suptitle(f"Scale {ires} | Direction {idir} | Angle {iang}")
            fig.tight_layout()


# # %%
# # Curvelet Coefficients: Real and Imaginary
# # #########################################
# opts["cmap"] = "gray"
# for ires, _ in enumerate(coeffs):
#     for idir, _ in enumerate(coeffs[ires]):
#         for iang, _ in enumerate(coeffs[ires][idir]):
#             z = coeffs[ires][idir][iang]
#             opts["vmax"] = np.abs(z).max()
#             opts["vmin"] = -opts["vmax"]
#             fig, axs = plt.subplots(1, 2, figsize=(8, 3))
#             for ax, img in zip(axs.ravel(), [z.real, z.imag]):
#                 im = ax.imshow(img.T, **opts)
#                 _, cb = create_colorbar(im=im, ax=ax)
#                 fmt = ticker.FuncFormatter(lambda x, _: f"{x:.0e}")
#                 cb.ax.yaxis.set_major_formatter(fmt)
#             axs[0].set(title="Real")
#             axs[1].set(title="Imaginary")
#             fig.suptitle(f"Scale {ires} | Direction {idir} | Angle {iang}")
#             fig.tight_layout()
# %%
# Curvelet Coefficients: Concatenated
# #########################################

# coeffs_vec = C.vect(coeffs=coeffs)
# coeffs_img = []
# for i in range(0, len(coeffs_vec) - zone_plate.size, zone_plate.size):
#     coeffs_img.append(coeffs_vec[i : i + zone_plate.size].reshape(1, *zone_plate.shape))
# if len(coeffs_vec) % zone_plate.size != 0:
#     coeffs_img.append(np.zeros((1, *zone_plate.shape), dtype=coeffs_vec.dtype))
#     coeffs_img[-1].flat[
#         len(coeffs_vec) - zone_plate.size : len(coeffs_vec)
#     ] = coeffs_vec[len(coeffs_vec) - zone_plate.size :]
# coeffs_img = np.concatenate(coeffs_img, axis=0)

_coeffs_vec = []
for c in coeffs[:2]:
    for d in c:
        for a in d:
            _coeffs_vec.append(a.ravel())
coeffs_vec = np.concatenate(_coeffs_vec)

coeffs_img = np.zeros((1, *zone_plate.shape), dtype=coeffs_vec.dtype)
coeffs_img.flat[: len(coeffs_vec)] = coeffs_vec[: len(coeffs_vec)]
# %%
fig, ax = plt.subplots(figsize=(12, 12))
opts["vmax"] = np.abs(coeffs_vec).max()
opts["vmin"] = 0
opts["cmap"] = "gray"
opts["aspect"] = "auto"
im = ax.imshow(np.abs(np.concatenate(coeffs_img, axis=0))[:50, :].T, **opts)
_, cb = create_colorbar(im=im, ax=ax)
fmt = ticker.FuncFormatter(lambda x, _: f"{x:.0e}")
cb.ax.yaxis.set_major_formatter(fmt)



from __future__ import annotations

from collections import deque
from collections.abc import Iterable
from typing import Any, Callable, TypeVar

import numpy as np
from numpy.fft import fft2, fftfreq, fftshift
from numpy.typing import NDArray

T = TypeVar("T")
U = TypeVar("U")
D = TypeVar("D", bound=np.generic)


def apply_along_wedges(
    c_struct: list[list[list[U]]],
    fun: Callable[
        [
            U,
            int,  # wedge index
            int,  # direction index
            int,  # scale index
            int,  # nwedges
            int,  # ndirections
            int,  # nscales
        ],
        T,
    ],
) -> list[list[list[T]]]:
    r"""Apply functions across a UDCT curvelet-like structure

    Parameters
    ----------
    c_struct : list[list[list[U]]]
        Input curvelet structure. First index: scales, second: directions, third: wedges.
    fun : Callable[[U, int, int, int, int, int, int], T]
        Function to apply to each item in the structure. The function's arguments
        are respectively: item, wedge index, direction index, scale index, number
        of wedges in direction, number of directions in scale, number of scales.

    Returns
    -------
    list[list[list[T]]]
        Result of applying the function to each item.

    Examples
    --------
    >>> import numpy as np
    >>> from curvelets.numpy import SimpleUDCT
    >>> from curvelets.utils import apply_along_wedges
    >>> x = np.zeros((32, 32))
    >>> C = SimpleUDCT(x.shape, nscales=3, nbands_per_direction=3)
    >>> y = C.forward(x)
    >>> apply_along_wedges(y, lambda w, *_: w.shape)
    [[[(16, 16)]],
     [[(8, 8), (8, 8), (8, 8)], [(8, 8), (8, 8), (8, 8)]],
     [[(16, 8), (16, 8), (16, 8), (16, 8), (16, 8), (16, 8)],
      [(8, 16), (8, 16), (8, 16), (8, 16), (8, 16), (8, 16)]]]

    """
    mapped_struct: list[list[list[T]]] = []
    for iscale, c_angles in enumerate(c_struct):
        tmp_scale = []
        for idir, c_dir in enumerate(c_angles):
            tmp_dir = []
            for iwedge, c_wedge in enumerate(c_dir):
                out = fun(
                    c_wedge,
                    iwedge,
                    idir,
                    iscale,
                    len(c_dir),
                    len(c_angles),
                    len(c_struct),
                )
                tmp_dir.append(out)
            tmp_scale.append(tmp_dir)
        mapped_struct.append(tmp_scale)
    return mapped_struct


def array_split_nd(ary: NDArray[D], *args: int) -> list[Any]:
    r"""Split an array into multiple sub-arrays recursively, possibly unevenly.

    Parameters
    ----------
    ary : :obj:`NDArray[D] <numpy.typing.NDArray>`
        Input array.
    args : :obj:`int`, optional
        Number of splits for each axis of ``ary``.
        Axis 0 will be split into ``args[0]`` subarrays, axis 1 will be
        into ``args[1]`` subarrays, etc. An axis of length
        ``l = ary.shape[axis]`` that should be split into ``n = args[axis]``
        sections, will return ``l % n`` sub-arrays of size ``l//n + 1``
        and the rest of size ``l//n``.

    Returns
    -------
    list[Any]
        Recursive lists of lists of :obj:`NDArray[D] <numpy.typing.NDArray>`.
        The number of depth of nesting is equivalent to the number arguments in args.

    See Also
    --------
    :obj:`numpy.array_split` : Split an array into multiple sub-arrays.

    Examples
    --------
    >>> from curvelets.utils import array_split_nd
    >>> ary = np.outer(1 + np.arange(2), 2 + np.arange(3))
    array([[2, 3, 4],
           [4, 6, 8]])
    >>> array_split_nd(ary, 2, 3)
    [[array([[2]]), array([[3]]), array([[4]])],
     [array([[4]]), array([[6]]), array([[8]])]]
    >>> ary = np.outer(np.arange(3), np.arange(5))
    >>> array_split_nd(ary, 2, 3)
    [[array([[0, 0],
             [0, 1]]),
      array([[0, 0],
             [2, 3]]),
      array([[0],
             [4]])],
     [array([[0, 2]]), array([[4, 6]]), array([[8]])]]
    """

    axis = ary.ndim - len(args)
    split = np.array_split(ary, args[0], axis=axis)
    if len(args) == 1:
        return split  # type: ignore[no-any-return]
    return [array_split_nd(s, *args[1:]) for s in split]


def deepflatten(lst: list[Any]) -> Iterable[Any]:
    r"""Flatten arbitrarily nested lists

    Parameters
    ----------
    lst : list[Any]
        Nested list of lists.

    Yields
    ------
    Iterator[Iterable[Any]]
        Elements in ``lst`` in order.

    Examples
    --------
    >>> from curvelets.utils import deepflatten
    >>> list(deepflatten([[[[[[1],2,[[3,4]]]]]]]))
    [1, 2, 3, 4]
    """
    q: Any = deque()
    for item in lst:
        if isinstance(item, list):
            q.extendleft(reversed(item))
        else:
            q.appendleft(item)
        while q:
            elem = q.popleft()
            if isinstance(elem, list):
                q.extendleft(reversed(elem))
            else:
                yield elem


def ndargmax(ary: NDArray[D]) -> tuple[np.intp, ...]:
    r"""N-dimensional argmax of array.

    Parameters
    ----------
    ary : :obj:`NDArray <numpy.typing.NDArray>`
        Input array.

    Returns
    -------
    :obj:`tuple` [:obj:`numpy.intp`, ...]
        N-dimensional index of the maximum of ``ary``.

    Examples
    --------
    >>> import numpy as np
    >>> from curvelets.utils import ndargmax
    >>> x = np.zeros((10, 10, 10))
    >>> x[1, 1, 1] = 1.0
    >>> ndargmax(x)
    (1, 1, 1)

    """
    return np.unravel_index(ary.argmax(), ary.shape)  # type: ignore[no-any-return]


def normal_vector_field(
    data: NDArray[np.generic], rows: int, cols: int, dx: float = 1, dy: float = 1
) -> NDArray[np.float64]:
    r"""Creates equally spaced vectors normal to the data.

    Parameters
    ----------
    data : :obj:`NDArray <numpy.typing.NDArray>`
        Input data. Plot with :obj:`plt.imshow <matplotlib.pyplot.imshow>` ``(data.T)``.
    rows : int
        Number of rows.
    cols : int
        Number of columns.
    dx : float, optional
        Spacing in the ``axis=0`` direction, by default 1
    dy : float, optional
        Spacing in the ``axis=1`` direction, by default 1

    Returns
    -------
    :obj:`NDArray <numpy.typing.NDArray>` [:obj:`numpy.float64`]
        Normal vectors shaped ``(nrows, ncols, 2)``
    """
    kvecs: NDArray[np.float64] = np.empty((rows, cols, 2), dtype=float)
    d_split: list[list[NDArray[D]]] = array_split_nd(data.T, rows, cols)

    for irow in range(kvecs.shape[0]):
        for icol in range(kvecs.shape[1]):
            d_loc = d_split[irow][icol].T
            d_k_loc = fftshift(fft2(d_loc))
            kx_loc = fftshift(fftfreq(d_loc.shape[0], d=dx))
            kz_loc = fftshift(fftfreq(d_loc.shape[1], d=dy))

            # Use top quadrants of f-k spectrum
            top_quadrant = kz_loc > 0
            kx_locmax, kz_locmax = ndargmax(np.abs(d_k_loc[:, top_quadrant]))  # pylint: disable=unbalanced-tuple-unpacking

            k = np.array([kx_loc[kx_locmax], kz_loc[top_quadrant][kz_locmax]])
            kvecs[irow, icol, :] = k / np.linalg.norm(k)
    return kvecs



from __future__ import annotations

import numpy as np

from ._circshift import circshift


# flip M-D matrix function ------------------------------------------------
def fftflip(F, axis):
    Fc = F
    dim = F.ndim
    shiftvec = np.zeros((dim,), dtype=int)
    shiftvec[axis] = 1
    Fc = np.flip(F, axis)
    Fc = circshift(Fc, shiftvec)
    return Fc



from __future__ import annotations

__all__ = ["udctmdwin"]
from itertools import combinations
from typing import Any

# import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt

from .typing import UDCTWindows
from .utils import (
    ParamUDCT,
    adapt_grid,
    angle_fun,
    angle_kron,
    circshift,
    fftflip,
    from_sparse_new,
    fun_meyer,
    to_sparse_new,
)


def _create_bandpass_windows(
    nscales: int, shape: tuple[int, ...], r: tuple[float, float, float, float]
) -> tuple[dict[int, np.ndarray], dict[int, np.ndarray]]:
    dim = len(shape)
    shape_grid: dict[int, np.ndarray] = {}
    meyers: dict[tuple[int, int], np.ndarray] = {}
    for idim in range(dim):
        # Don't take the np.pi out of the linspace
        shape_grid[idim] = np.linspace(
            -1.5 * np.pi, 0.5 * np.pi, shape[idim], endpoint=False
        )

        params = np.array([-2, -1, *r[:2]])
        abs_shape_grid = np.abs(shape_grid[idim])
        meyers[(nscales, idim)] = fun_meyer(abs_shape_grid, *params)
        if nscales == 1:
            meyers[(nscales, idim)] += fun_meyer(
                np.abs(shape_grid[idim] + 2 * np.pi), *params
            )
        params[2:] = r[2:]
        meyers[(nscales + 1, idim)] = fun_meyer(abs_shape_grid, *params)

        for jn in range(nscales - 1, 0, -1):
            params[2:] = r[:2]
            params[2:] /= 2 ** (nscales - jn)
            meyers[(jn, idim)] = fun_meyer(abs_shape_grid, *params)

    bandpasses: dict[int, np.ndarray] = {}
    for jn in range(nscales, 0, -1):
        lo = np.array([1.0])
        hi = np.array([1.0])
        for idim in range(dim - 1, -1, -1):
            lo = np.kron(meyers[(jn, idim)], lo)
            hi = np.kron(meyers[(jn + 1, idim)], hi)
        lo_nd = lo.reshape(*shape)
        hi_nd = hi.reshape(*shape)
        bp_nd = hi_nd - lo_nd
        bp_nd[bp_nd < 0] = 0
        bandpasses[jn] = bp_nd
    bandpasses[0] = lo_nd
    return shape_grid, bandpasses


def _nchoosek(n: Any, k: Any) -> np.ndarray:
    return np.asarray(list(combinations(n, k)))


def _create_mdirs(dim: int, res: int) -> list[np.ndarray]:
    # Mdir is dimension of need to calculate angle function on each
    # hyperpyramid
    return [
        np.c_[[np.r_[np.arange(idim), np.arange(idim + 1, dim)] for idim in range(dim)]]
        for ires in range(res)
    ]
    # Mdirs: dict[int, np.ndarray] = {}
    # for ires in range(res):
    #     Mdirs[ires] = np.zeros((dim, dim - 1), dtype=int)
    #     for idim in range(dim):
    #         Mdirs[ires][idim, :] = np.r_[range(idim), range(idim + 1, dim)]

    # return Mdirs


def _create_angle_info(
    Sgrid: dict[int, np.ndarray], dim: int, res: int, cfg: np.ndarray, alpha: float
) -> tuple[
    dict[int, dict[tuple[int, int], np.ndarray]],
    dict[int, dict[tuple[int, int], np.ndarray]],
]:
    # every combination of 2 dimension out of 1:dim
    mperms = _nchoosek(np.arange(dim), 2)
    Mgrid: dict[tuple[int, int], np.ndarray] = {}
    for ind, perm in enumerate(mperms):
        out = adapt_grid(Sgrid[perm[0]], Sgrid[perm[1]])
        Mgrid[(ind, 0)] = out[0]
        Mgrid[(ind, 1)] = out[1]

    # gather angle function for each pyramid
    Mangs: dict[int, dict[tuple[int, int], np.ndarray]] = {}
    Minds: dict[int, dict[tuple[int, int], np.ndarray]] = {}
    for ires in range(res):
        Mangs[ires] = {}
        Minds[ires] = {}
        # for each resolution
        for idim in range(dim):
            # for each pyramid in resolution res
            cnt = 0
            # cnt is number of angle function required for each pyramid
            # now loop through mperms
            for ihyp in range(mperms.shape[0]):
                for idir in range(mperms.shape[1]):
                    if mperms[ihyp, idir] == idim:
                        Mangs[ires][(idim, cnt)] = angle_fun(
                            Mgrid[(ihyp, idir)],
                            idir + 1,
                            cfg[ires, mperms[ihyp, 1 - idir]],
                            alpha,
                        )
                        Minds[ires][(idim, cnt)] = mperms[ihyp, :] + 1
                        cnt += 1
    return Mangs, Minds


def _inplace_normalize_windows(
    udctwin: UDCTWindows, size: tuple[int, ...], dim: int, res: int
) -> None:
    sumw2 = np.zeros(size)
    idx, val = from_sparse_new(udctwin[0][0][0])
    sumw2.flat[idx] += val**2
    for ires in range(1, res + 1):
        for idir in range(dim):
            for iang in range(len(udctwin[ires][idir])):
                tmpw = np.zeros(size)
                idx, val = from_sparse_new(udctwin[ires][idir][iang])
                tmpw.flat[idx] += val**2
                sumw2 += tmpw
                tmpw = fftflip(tmpw, idir)
                sumw2 += tmpw

    sumw2 = np.sqrt(sumw2)
    idx, val = from_sparse_new(udctwin[0][0][0])
    val /= sumw2.ravel()[idx]
    for ires in range(1, res + 1):
        for idir in range(dim):
            for iang in range(len(udctwin[ires][idir])):
                idx, val = from_sparse_new(udctwin[ires][idir][iang])
                val /= sumw2.ravel()[idx]


def _calculate_decimation_ratios_with_lowest(
    res: int, dim: int, cfg: np.ndarray, Mdirs: list[np.ndarray]
) -> list[npt.NDArray[np.int_]]:
    decimation_ratio: list[npt.NDArray[np.int_]] = [
        np.full((1, dim), fill_value=2 ** (res - 1), dtype=int)
    ]
    for ires in range(1, res + 1):
        decimation_ratio.append(
            np.full((dim, dim), fill_value=2.0 ** (res - ires + 1), dtype=int)
        )
        for i0 in range(dim):
            i1s = Mdirs[ires - 1][i0, :]
            decimation_ratio[ires][i0, i1s] = (
                2 * cfg[ires - 1, i1s] * 2 ** (res - ires) // 3
            )
    return decimation_ratio


def _inplace_sort_windows(
    udctwin: UDCTWindows, indices: dict[int, dict[int, np.ndarray]], res: int, dim: int
) -> None:
    for ires in range(1, res + 1):
        for idim in range(dim):
            mlist = indices[ires][idim]

            # # Approach 1: Create a structured array and then sort by fields
            # struct = [(f"x{i}", "<i8") for i in range(mlist.shape[1])]
            # ix = np.argsort(
            #     np.array([tuple(m) for m in mlist], dtype=struct),
            #     order=tuple(t[0] for t in struct),
            # )
            #
            # Approach 2: Create a 1D array then sort that array
            m = mlist.max() + 1
            ix = np.argsort(
                sum(
                    m**i2 * mlist[:, i1]
                    for i1, i2 in enumerate(range(mlist.shape[1] - 1, -1, -1))
                )
            )

            indices[ires][idim] = mlist[ix]
            udctwin[ires][idim] = [udctwin[ires][idim][idx] for idx in ix]


def udctmdwin(
    param_udct: ParamUDCT,
) -> tuple[UDCTWindows, list[npt.NDArray[np.int_]], dict[int, dict[int, np.ndarray]]]:
    Sgrid, F2d = _create_bandpass_windows(
        nscales=param_udct.res, shape=param_udct.size, r=param_udct.r
    )
    Winlow = circshift(np.sqrt(F2d[0]), tuple(s // 4 for s in param_udct.size))

    # convert to sparse format
    udctwin: UDCTWindows = []
    udctwin.append([])
    udctwin[0].append([])
    udctwin[0][0] = [to_sparse_new(Winlow, param_udct.winthresh)]

    # `indices` gets stored as `param_udct.ind` in the original.
    indices: dict[int, dict[int, np.ndarray]] = {}
    indices[0] = {}
    indices[0][0] = np.zeros((1, 1), dtype=int)
    Mdirs = _create_mdirs(dim=param_udct.dim, res=param_udct.res)
    Mangs, Minds = _create_angle_info(
        Sgrid,
        dim=param_udct.dim,
        res=param_udct.res,
        cfg=param_udct.cfg,
        alpha=param_udct.alpha,
    )

    # decimation ratio for each band
    decimation_ratio = _calculate_decimation_ratios_with_lowest(
        res=param_udct.res, dim=param_udct.dim, cfg=param_udct.cfg, Mdirs=Mdirs
    )

    # Mang is 1-d angle function for each hyper pyramid (row) and each angle
    # dimension (column)
    for ires in range(1, param_udct.res + 1):  # pylint: disable=too-many-nested-blocks
        # for each resolution
        udctwin.append([])
        indices[ires] = {}
        for idim in range(param_udct.dim):
            udctwin[ires].append([])
            # for each hyperpyramid
            i1_ang = np.arange(len(Mangs[ires - 1][(idim, 0)]))[:, None]
            for i2 in range(1, param_udct.dim - 1):
                ln = len(Mangs[ires - 1][(idim, i2)])
                tmp2 = np.arange(len(Mangs[ires - 1][(idim, i2)]))[:, None]
                tmp3 = np.kron(i1_ang, np.ones((ln, 1), dtype=int))
                tmp4 = np.kron(np.ones((i1_ang.shape[0], 1), dtype=int), tmp2)
                i1_ang = np.c_[tmp3, tmp4]
            lent = i1_ang.shape[0]
            ang_inmax = param_udct.cfg[ires - 1, Mdirs[ires - 1][idim, :]]
            # lent is the smallest number of windows need to calculated on each
            # pyramid
            # ang_inmax is M-1 vector contain number of angle function per each
            # dimension of the hyperpyramid
            for i3 in range(lent):
                # for each calculated windows function, estimated all the other
                # flipped window functions
                win: npt.NDArray[np.floating] = np.ones(param_udct.size, dtype=float)
                for i4 in range(param_udct.dim - 1):
                    idx = i1_ang.reshape(len(i1_ang), -1)[i3, i4]
                    tmp1 = Mangs[ires - 1][(idim, i4)][idx]
                    tmp2 = Minds[ires - 1][(idim, i4)]
                    afun2 = angle_kron(tmp1, tmp2, param_udct)
                    win *= afun2
                win *= F2d[ires]
                win = np.sqrt(circshift(win, tuple(s // 4 for s in param_udct.size)))

                # first windows function
                angle_functions = []
                angle_functions.append(win)

                # index of current angle
                i2_ang = i1_ang[i3 : i3 + 1, :] + 1

                # all possible flip along different dimension
                for i5 in range(param_udct.dim - 2, -1, -1):
                    for i6 in range(i2_ang.shape[0]):
                        if 2 * i2_ang[i6, i5] <= ang_inmax[i5]:
                            i2_ang_tmp = i2_ang[i6 : i6 + 1, :].copy()
                            i2_ang_tmp[0, i5] = ang_inmax[i5] + 1 - i2_ang[i6, i5]
                            i2_ang = np.r_[i2_ang, i2_ang_tmp]
                            win = fftflip(
                                angle_functions[i6], Mdirs[ires - 1][idim, i5]
                            )
                            angle_functions.append(win)
                i2_ang -= 1  # Adjust so that `indices` is 0-based
                angle_functions = np.c_[angle_functions]

                if i3 == 0:
                    ang_ind = i2_ang
                    for i7 in range(ang_ind.shape[0]):
                        udctwin[ires][idim].append(
                            to_sparse_new(angle_functions[i7], param_udct.winthresh)
                        )
                else:
                    inold = ang_ind.shape[0]
                    ang_ind = np.concatenate((ang_ind, i2_ang), axis=0)
                    innew = ang_ind.shape[0]
                    for i7 in range(inold, innew):
                        udctwin[ires][idim].append(
                            to_sparse_new(
                                angle_functions[i7 - inold], param_udct.winthresh
                            )
                        )
                    indices[ires][idim] = ang_ind.copy()

    # Normalization
    _inplace_normalize_windows(
        udctwin, size=param_udct.size, dim=param_udct.dim, res=param_udct.res
    )

    # sort the window
    _inplace_sort_windows(
        udctwin=udctwin, indices=indices, res=param_udct.res, dim=param_udct.dim
    )

    return udctwin, decimation_ratio, indices



from __future__ import annotations

import numpy as np


def adapt_grid(S1: np.ndarray, S2: np.ndarray):
    x1, x2 = np.meshgrid(S2, S1)

    t1 = np.zeros_like(x1, dtype=float)
    ind = (x1 != 0) & (np.abs(x2) <= np.abs(x1))
    t1[ind] = -x2[ind] / x1[ind]

    t2 = np.zeros_like(x1, dtype=float)
    ind = (x2 != 0) & (np.abs(x1) < np.abs(x2))
    t2[ind] = x1[ind] / x2[ind]

    t3 = t2.copy()
    t3[t2 < 0] = t2[t2 < 0] + 2
    t3[t2 > 0] = t2[t2 > 0] - 2

    M1 = t1 + t3
    M1[x1 >= 0] = -2

    t1 = np.zeros_like(x1, dtype=float)
    ind = (x2 != 0) & (abs(x1) <= abs(x2))
    t1[ind] = -x1[ind] / x2[ind]

    t2 = np.zeros_like(x1, dtype=float)
    ind = (x1 != 0) & (abs(x2) < abs(x1))
    t2[ind] = x2[ind] / x1[ind]

    t3 = t2.copy()
    t3[t2 < 0] = t2[t2 < 0] + 2
    t3[t2 > 0] = t2[t2 > 0] - 2

    M2 = t1 + t3
    M2[x2 >= 0] = -2

    return M2, M1



from __future__ import annotations

import numpy as np
import pytest

import curvelets.numpy as udct
import curvelets.reference as udct_ref
from curvelets.numpy.utils import from_sparse, from_sparse_new


@pytest.mark.parametrize("dim", list(range(2, 5)))
def test_compare_with_reference(dim):
    rng = np.random.default_rng()

    opts = [16, 32, 64, 128, 256]
    if dim == 3:
        opts = opts[:2]
    elif dim >= 4:
        opts = opts[:1]
    size = rng.choice(opts, size=dim, replace=True)
    cfg = (
        np.array([[3, 3], [6, 6], [12, 6]])
        if dim == 2
        else np.c_[np.ones((dim,)) * 3, np.ones((dim,)) * 6].T
    )
    alpha = 0.3 * rng.uniform(size=1).item()
    r: tuple[float, float, float, float] = tuple(
        np.pi * np.array([1.0, 2.0, 2.0, 4.0]) / 3
    )
    winthresh = 10.0 ** (-rng.integers(low=4, high=6, size=1).item())

    my_udct = udct.UDCT(shape=size, cfg=cfg, alpha=alpha, r=r, winthresh=winthresh)
    param_ref = udct_ref.ParamUDCT(
        dim=dim, size=size, cfg=cfg, alpha=alpha, r=r, winthresh=winthresh
    )

    udctwin = my_udct.windows
    udctwin_ref = udct_ref.udctmdwin(param_ref)
    assert param_ref.ind is not None

    rdtype = udctwin[0][0][0][1].real.dtype
    cdtype = (np.ones(1, dtype=rdtype) + 1j * np.ones(1, dtype=rdtype)).dtype
    win = np.zeros(param_ref.size, dtype=cdtype)
    win_ref = np.zeros(param_ref.size, dtype=cdtype)

    np.testing.assert_array_equal(
        1 + np.arange(len(udctwin)), np.asarray(list(udctwin_ref.keys()))
    )
    for res in udctwin_ref:
        np.testing.assert_array_equal(
            1 + np.arange(len(udctwin[res - 1])),
            np.asarray(list(udctwin_ref[res].keys())),
        )
        for dir in udctwin_ref[res]:
            if res == 1 and dir == 1:
                np.testing.assert_array_equal(
                    my_udct.indices[0][0], param_ref.ind[1][1]
                )
                idx, val = from_sparse_new(udctwin[0][0][0])
                win[...] = 0  # Reset
                win.flat[idx] = val  # Fill

                idx, val = from_sparse(udctwin_ref[1][1])
                win_ref[...] = 0  # Reset
                win_ref.T.flat[idx] = val  # Fill
                np.testing.assert_allclose(win, win_ref, rtol=1e-14)

            else:
                np.testing.assert_array_equal(
                    1 + my_udct.indices[res - 1][dir - 1], param_ref.ind[res][dir]
                )
                np.testing.assert_array_equal(
                    1 + np.arange(len(udctwin[res - 1][dir - 1])),
                    np.asarray(list(udctwin_ref[res][dir].keys())),
                )
                for ang in udctwin_ref[res][dir]:
                    idx, val = from_sparse_new(udctwin[res - 1][dir - 1][ang - 1])
                    win[...] = 0  # Reset
                    win.flat[idx] = val  # Fill

                    idx, val = from_sparse(udctwin_ref[res][dir][ang])
                    win_ref[...] = 0  # Reset
                    win_ref.T.flat[idx] = val  # Fill
                    np.testing.assert_allclose(win, win_ref, rtol=1e-14)

    im = rng.normal(size=size)
    coeffs = my_udct.forward(im)
    coeffs_ref = udct_ref.udctmddec(im, param_udct=param_ref, udctwin=udctwin_ref)

    np.testing.assert_array_equal(
        1 + np.arange(len(coeffs)), np.asarray(list(coeffs_ref.keys()))
    )
    for res in coeffs_ref:
        np.testing.assert_array_equal(
            1 + np.arange(len(coeffs[res - 1])),
            np.asarray(list(coeffs_ref[res].keys())),
        )
        for dir in coeffs_ref[res]:
            if res == 1 and dir == 1:
                np.testing.assert_allclose(
                    coeffs[res - 1][dir - 1][0], coeffs_ref[res][dir], rtol=1e-14
                )
            else:
                np.testing.assert_array_equal(
                    1 + np.arange(len(coeffs[res - 1][dir - 1])),
                    np.asarray(list(coeffs_ref[res][dir].keys())),
                )
                for ang in coeffs_ref[res][dir]:
                    np.testing.assert_allclose(
                        coeffs[res - 1][dir - 1][ang - 1],
                        coeffs_ref[res][dir][ang],
                        rtol=1e-14,
                    )
    im2 = my_udct.backward(coeffs)
    im2_ref = udct_ref.udctmdrec(coeffs_ref, param_udct=param_ref, udctwin=udctwin_ref)
    np.testing.assert_allclose(im2, im2_ref, rtol=1e-14)


@pytest.mark.parametrize("dim", list(range(2, 4)))
def test_round_trip_absolute(dim):
    rng = np.random.default_rng()

    # For these specific parameters, we can guarantee an absolute precision of
    # approximately 1e-4
    size: tuple[int, ...]
    if dim == 2:
        size = (256, 256)
    elif dim == 3:
        size = tuple(4 * np.array([32, 32, 32]))
    elif dim == 3:
        size = (16, 16, 16, 16)
    cfg = (
        np.array([[3, 3], [6, 6], [12, 6]])
        if dim == 2
        else np.c_[np.ones((dim,)) * 3, np.ones((dim,)) * 6].T
    )
    alpha = 0.15
    r = tuple(np.pi * np.array([1.0, 2.0, 2.0, 4.0]) / 3)
    winthresh = 1e-5

    my_udct = udct.UDCT(shape=size, cfg=cfg, alpha=alpha, r=r, winthresh=winthresh)
    im = rng.normal(size=size)
    coeffs = my_udct.forward(im)
    im2 = my_udct.backward(coeffs)

    # try:
    #     import matplotlib.pyplot as plt

    #     idx = [np.random.choice(s) for s in im.shape]
    #     for i in np.random.choice(im.ndim, 2, replace=False):
    #         idx[i] = slice(None)
    #     idx = tuple(idx)

    #     fig, axs = plt.subplots(1, 3, figsize=(12, 3))
    #     img = axs[0].imshow(im[idx])
    #     fig.colorbar(img)
    #     img = axs[1].imshow(im2[idx])
    #     fig.colorbar(img)
    #     img = axs[2].imshow((im - im2)[idx])
    #     fig.colorbar(img)
    #     fig.suptitle(f"Error: {np.abs(im - im2).max()}")
    #     fig.tight_layout()
    #     plt.show()
    # except ImportError:
    #     pass

    np.testing.assert_allclose(im, im2, atol=1e-4)


@pytest.mark.parametrize("dim", list(range(2, 5)))
def test_round_trip_rel(dim):
    rng = np.random.default_rng()

    # For random parameters in the range below, we can guarantee an relative precision of
    # approximately 0.5% of the maximum amplitude in the original image.
    opts = [16, 32, 64, 128, 256]
    if dim == 3:
        opts = opts[:3]
    elif dim >= 4:
        opts = opts[:1]
    size = rng.choice(opts, size=dim, replace=True)
    cfg = (
        np.array([[3, 3], [6, 6], [12, 6]])
        if dim == 2
        else np.c_[np.ones((dim,)) * 3, np.ones((dim,)) * 6].T
    )
    alpha = 0.3 * rng.uniform(size=1)
    r = np.pi * np.array([1.0, 2.0, 2.0, 4.0]) / 3
    winthresh = 10.0 ** (-rng.integers(low=4, high=6, size=1))

    my_udct = udct.UDCT(shape=size, cfg=cfg, alpha=alpha, r=r, winthresh=winthresh)
    im = rng.normal(size=size)
    coeffs = my_udct.forward(im)
    im2 = my_udct.backward(coeffs)

    # try:
    #     import matplotlib.pyplot as plt

    #     idx = [np.random.choice(s) for s in im.shape]
    #     for i in np.random.choice(im.ndim, 2, replace=False):
    #         idx[i] = slice(None)
    #     idx = tuple(idx)

    #     fig, axs = plt.subplots(1, 3, figsize=(12, 3))
    #     img = axs[0].imshow(im[idx])
    #     fig.colorbar(img)
    #     img = axs[1].imshow(im2[idx])
    #     fig.colorbar(img)
    #     img = axs[2].imshow((im - im2)[idx])
    #     fig.colorbar(img)
    #     fig.suptitle(f"Error: {np.abs(im - im2).max()}")
    #     fig.tight_layout()
    #     plt.show()
    # except ImportError:
    #     pass

    np.testing.assert_allclose(im, im2, atol=0.005 * im.max())


@pytest.mark.parametrize("dim", list(range(2, 5)))
def test_vect_struct(dim):
    rng = np.random.default_rng()

    opts = [16, 32, 64, 128, 256]
    if dim == 3:
        opts = opts[:3]
    elif dim >= 4:
        opts = opts[:1]
    shape: tuple[int, ...] = tuple(rng.choice(opts, size=dim, replace=True))
    cfg = (
        np.array([[3, 3], [6, 6], [12, 6]])
        if dim == 2
        else np.c_[np.ones((dim,)) * 3, np.ones((dim,)) * 6].T
    )
    alpha = 0.3 * rng.uniform(size=1).item()
    r: tuple[float, float, float, float] = tuple(
        np.pi * np.array([1.0, 2.0, 2.0, 4.0]) / 3
    )
    winthresh = 10.0 ** (-rng.integers(low=4, high=6, size=1).item())

    C = udct.UDCT(shape=shape, cfg=cfg, alpha=alpha, r=r, winthresh=winthresh)
    im = rng.normal(size=shape)
    c1 = C.forward(im)

    c2 = C.struct(C.vect(c1))
    for ires, _ in enumerate(c1):
        for idir, _ in enumerate(c1[ires]):
            for iang, _ in enumerate(c1[ires][idir]):
                np.testing.assert_allclose(c1[ires][idir][iang], c2[ires][idir][iang])


@pytest.mark.parametrize("dim", list(range(2, 4)))
def test_round_trip_absolute_transp(dim):
    rng = np.random.default_rng()

    # For these specific parameters, we can guarantee an absolute precision of
    # approximately 1e-4
    size: tuple[int, ...]
    if dim == 2:
        size = (256, 256)
    elif dim == 3:
        size = tuple(4 * np.array([32, 32, 32]))
    elif dim == 3:
        size = (16, 16, 16, 16)
    cfg = (
        np.array([[3, 3], [6, 6], [12, 6]])
        if dim == 2
        else np.c_[np.ones((dim,)) * 3, np.ones((dim,)) * 6].T
    )
    alpha = 0.15
    r = tuple(np.pi * np.array([1.0, 2.0, 2.0, 4.0]) / 3)
    winthresh = 1e-5

    my_udct = udct.UDCT(
        shape=size, cfg=cfg, alpha=alpha, r=r, winthresh=winthresh, transpose=True
    )
    im = rng.normal(size=size)
    coeffs = my_udct.forward(im)
    im2 = my_udct.backward(coeffs)

    np.testing.assert_allclose(im, im2, atol=1e-4)


@pytest.mark.parametrize("dim", list(range(2, 5)))
def test_round_trip_rel_transp(dim):
    rng = np.random.default_rng()

    # For random parameters in the range below, we can guarantee an relative precision of
    # approximately 0.5% of the maximum amplitude in the original image.
    opts = [16, 32, 64, 128, 256]
    if dim == 3:
        opts = opts[:3]
    elif dim >= 4:
        opts = opts[:1]
    size = rng.choice(opts, size=dim, replace=True)
    cfg = (
        np.array([[3, 3], [6, 6], [12, 6]])
        if dim == 2
        else np.c_[np.ones((dim,)) * 3, np.ones((dim,)) * 6].T
    )
    alpha = 0.3 * rng.uniform(size=1)
    r = np.pi * np.array([1.0, 2.0, 2.0, 4.0]) / 3
    winthresh = 10.0 ** (-rng.integers(low=4, high=6, size=1))

    my_udct = udct.UDCT(
        shape=size, cfg=cfg, alpha=alpha, r=r, winthresh=winthresh, transpose=True
    )
    im = rng.normal(size=size)
    coeffs = my_udct.forward(im)
    im2 = my_udct.backward(coeffs)

    np.testing.assert_allclose(im, im2, atol=0.005 * im.max())


@pytest.mark.parametrize("dim", list(range(2, 5)))
def test_vect_struct_transp(dim):
    rng = np.random.default_rng()

    opts = [16, 32, 64, 128, 256]
    if dim == 3:
        opts = opts[:3]
    elif dim >= 4:
        opts = opts[:1]
    shape: tuple[int, ...] = tuple(rng.choice(opts, size=dim, replace=True))
    cfg = (
        np.array([[3, 3], [6, 6], [12, 6]])
        if dim == 2
        else np.c_[np.ones((dim,)) * 3, np.ones((dim,)) * 6].T
    )
    alpha = 0.3 * rng.uniform(size=1).item()
    r: tuple[float, float, float, float] = tuple(
        np.pi * np.array([1.0, 2.0, 2.0, 4.0]) / 3
    )
    winthresh = 10.0 ** (-rng.integers(low=4, high=6, size=1).item())

    C = udct.UDCT(
        shape=shape, cfg=cfg, alpha=alpha, r=r, winthresh=winthresh, transpose=True
    )
    im = rng.normal(size=shape)
    c1 = C.forward(im)

    c2 = C.struct(C.vect(c1))
    for ires, _ in enumerate(c1):
        for idir, _ in enumerate(c1[ires]):
            for iang, _ in enumerate(c1[ires][idir]):
                np.testing.assert_allclose(c1[ires][idir][iang], c2[ires][idir][iang])



r"""
Multiscale Direction Disks
==========================
This example shows how to use the UDCT curvelets transform to visualize
multiscale preferrential directions in an image. Inspired by
`Kymatio's Scattering disks <https://www.kymat.io/gallery_2d/plot_scattering_disk.html>`__.
"""

from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np

from curvelets.numpy import SimpleUDCT
from curvelets.plot import create_inset_axes_grid, overlay_arrows, overlay_disk
from curvelets.utils import apply_along_wedges, normal_vector_field

# %%
# Input Data
# ##########
inputfile = "../testdata/sigmoid.npz"
data = np.load(inputfile)
data = data["sigmoid"][:120, :64]
nx, nz = data.shape
dx, dz = 0.005, 0.004
x, z = np.arange(nx) * dx, np.arange(nz) * dz

###############################################################################
aspect = dz / dx
figsize_aspect = aspect * nz / nx
opts_space = {
    "extent": (x[0], x[-1], z[-1], z[0]),
    "cmap": "gray",
    "interpolation": "lanczos",
    "aspect": aspect,
}
vmax = 0.5 * np.max(np.abs(data))
fig, ax = plt.subplots(figsize=(8, figsize_aspect * 8))
ax.imshow(data.T, vmin=-vmax, vmax=vmax, **opts_space)
ax.set(xlabel="Position [km]", ylabel="Depth [km]", title="Data")

# %%
# UDCT
Cop = SimpleUDCT(data.shape, nscales=3, nbands_per_direction=3, transpose=True)
d_c = Cop.forward(data)

# %%
# Compute average normal vector. This vector indicates the direction normal to the structures
# of the image
kvecs = normal_vector_field(data, 1, 1)
kvecs *= 0.4 * min(x[-1] - x[0], z[-1] - z[0])

# %%
# Now we compute the average energy of each curvelet "wedge". We will plot this on a
# multiscale disk to show the distribution of energies along different directions of the various
# scales in the data.
energy_c = apply_along_wedges(d_c, lambda w, *_: np.sqrt((np.abs(w) ** 2).mean()))

# %%
fig, ax = plt.subplots(figsize=(12, figsize_aspect * 8))
ax.imshow(data.T, vmin=-vmax, vmax=vmax, **opts_space)
overlay_arrows(kvecs, ax, arrowprops={"edgecolor": "w", "facecolor": "k"})
ax_o = create_inset_axes_grid(ax, width=0.4, kwargs_inset_axes={"projection": "polar"})
overlay_disk(energy_c, ax=ax_o, vmin=0, cmap="turbo", linecolor="w", linewidth=5)
ax.set(xlabel="Position [km]", ylabel="Depth [km]", title="Data")



from __future__ import annotations

import sys
from dataclasses import dataclass, field

import numpy as np

from ._downsamp import downsamp
from ._upsamp import upsamp


@dataclass(**(dict(kw_only=True) if sys.version_info >= (3, 10) else {}))
class ParamUDCT:
    dim: int
    size: tuple[int, int] | tuple[int, int, int]
    cfg: tuple | np.ndarray  # last dimension  == dim
    alpha: float
    r: tuple[float, float, float, float]
    winthresh: float
    len: int = field(init=False)
    res: int = field(init=False)
    decim: np.ndarray = field(init=False)
    ind: dict[int, dict[int, np.ndarray]] | None = None
    dec: dict[int, np.ndarray] | None = None

    def __post_init__(self):
        self.len = np.prod(self.size)
        self.res = len(self.cfg)
        self.decim = 2 * (np.asarray(self.cfg, dtype=int) // 3)


def udctmddec(im, param_udct, udctwin):
    imf = np.fft.fftn(im)

    fband = np.zeros_like(imf)
    idx = udctwin[1][1][:, 0].astype(int) - 1
    val = udctwin[1][1][:, 1]
    fband.T.flat[idx] = imf.T.flat[idx] * val
    cband = np.fft.ifftn(fband)

    coeff = {}
    coeff[1] = {}
    decim = np.full((param_udct.dim,), fill_value=2 ** (param_udct.res - 1), dtype=int)
    coeff[1][1] = downsamp(cband, decim)
    norm = np.sqrt(
        np.prod(np.full((param_udct.dim,), fill_value=2 ** (param_udct.res - 1)))
    )
    coeff[1][1] *= norm

    for res in range(1, 1 + param_udct.res):
        coeff[res + 1] = {}
        for dir in range(1, 1 + param_udct.dim):
            coeff[res + 1][dir] = {}
            for ang in range(1, 1 + len(udctwin[res + 1][dir])):
                fband = np.zeros_like(imf)
                idx = udctwin[res + 1][dir][ang][:, 0].astype(int) - 1
                val = udctwin[res + 1][dir][ang][:, 1]
                fband.T.flat[idx] = imf.T.flat[idx] * val

                cband = np.fft.ifftn(fband)
                decim = param_udct.dec[res][dir - 1, :].astype(int)
                coeff[res + 1][dir][ang] = downsamp(cband, decim)
                coeff[res + 1][dir][ang] *= np.sqrt(
                    2 * np.prod(param_udct.dec[res][dir - 1, :])
                )
    return coeff


def udctmdrec(coeff, param_udct, udctwin):
    imf = np.zeros(param_udct.size, dtype=np.complex128)

    for res in range(1, 1 + param_udct.res):
        for dir in range(1, 1 + param_udct.dim):
            for ang in range(1, 1 + len(udctwin[res + 1][dir])):
                decim = param_udct.dec[res][dir - 1, :].astype(int)
                cband = upsamp(coeff[res + 1][dir][ang], decim)
                cband /= np.sqrt(2 * np.prod(param_udct.dec[res][dir - 1, :]))
                cband = np.prod(param_udct.dec[res][dir - 1, :]) * np.fft.fftn(cband)
                idx = udctwin[res + 1][dir][ang][:, 0].astype(int) - 1
                val = udctwin[res + 1][dir][ang][:, 1]
                imf.T.flat[idx] += cband.T.flat[idx] * val

    imfl = np.zeros(param_udct.size, dtype=np.complex128)
    decimlow = np.full(
        (param_udct.dim,), fill_value=2 ** (param_udct.res - 1), dtype=int
    )
    cband = upsamp(coeff[1][1], decimlow)
    cband = np.sqrt(np.prod(decimlow)) * np.fft.fftn(cband)
    idx = udctwin[1][1][:, 0].astype(int) - 1
    val = udctwin[1][1][:, 1]
    imfl.T.flat[idx] += cband.T.flat[idx] * val
    imf = 2 * imf + imfl
    im2 = np.fft.ifftn(imf).real
    return im2



from __future__ import annotations

__all__ = [
    "create_colorbar",
    "create_inset_axes_grid",
    "despine",
    "overlay_arrows",
    "overlay_disk",
]
import logging

from .._internal import MATPLOTLIB_ENABLED
from ._curvelet import overlay_disk

logger = logging.getLogger()

if MATPLOTLIB_ENABLED:
    from ._matplotlib import (
        create_colorbar,
        create_inset_axes_grid,
        despine,
        overlay_arrows,
    )
else:
    logger.warning("matplotlib is not installed, not all functions will be available")



from __future__ import annotations

import numpy as np


def upsamp(F, decim):
    assert F.ndim == len(decim)
    upsamp_shape = tuple(s * d for s, d in zip(F.shape, decim))
    D = np.zeros(upsamp_shape, dtype=F.dtype)
    D[tuple(slice(None, None, d) for d in decim)] = F[...]
    return D



from __future__ import annotations

from copy import deepcopy
from itertools import combinations

import numpy as np

from ._adapt_grid import adapt_grid
from ._angle_fun import angle_fun
from ._angle_kron import angle_kron
from ._circshift import circshift
from ._fftflip import fftflip
from ._fun_meyer import fun_meyer
from ._travel import travel
from .udct import ParamUDCT


def _to_sparse(arr, thresh):
    idx = np.argwhere(travel(arr) > thresh)
    return np.c_[idx + 1, travel(arr)[idx]]


def udctmdwin(param_udct: ParamUDCT) -> dict:
    Sgrid = {}
    f1d = {}
    for ind in range(1, param_udct.dim + 1):
        start = -1.5 * np.pi
        stop = 0.5 * np.pi  # - np.pi / (param_udct.size[ind - 1] / 2)
        # step = np.pi / (param_udct.size[ind - 1] / 2)
        # np.arange(start, stop + step, step)
        Sgrid[ind] = np.linspace(start, stop, param_udct.size[ind - 1], endpoint=False)

        params = np.array([-2, -1, *param_udct.r[:2]])
        f1d[(param_udct.res, ind)] = fun_meyer(np.abs(Sgrid[ind]), *params)
        if param_udct.res == 1:
            f1d[(param_udct.res, ind)] += fun_meyer(
                np.abs(Sgrid[ind] + 2 * np.pi), *params
            )
        params[2:] = param_udct.r[2:]
        f1d[(param_udct.res + 1, ind)] = fun_meyer(abs(Sgrid[ind]), *params)

        for jn in range(param_udct.res - 1, 0, -1):
            params[2:] = param_udct.r[:2]
            params[2:] /= 2 ** (param_udct.res - jn)
            f1d[(jn, ind)] = fun_meyer(abs(Sgrid[ind]), *params)

    F2d = {}
    for jn in range(param_udct.res, 0, -1):
        fltmp = np.array([1.0])
        fhtmp = np.array([1.0])
        for ind in range(param_udct.dim, 0, -1):
            fltmp = np.kron(fltmp.ravel(), f1d[(jn, ind)].ravel())
            fhtmp = np.kron(fhtmp.ravel(), f1d[(jn + 1, ind)].ravel())
        FL = fltmp.reshape(*param_udct.size[::-1]).T.astype(float).copy()
        FH = fhtmp.reshape(*param_udct.size[::-1]).T.astype(float).copy()
        FP = FH - FL
        FP[FP < 0] = 0
        F2d[jn + 1] = FP
    F2d[1] = FL.copy()

    Winlow = circshift(np.sqrt(F2d[1]), tuple(s // 4 for s in param_udct.size))
    # convert to sparse format
    udctwin = {}
    udctwin[1] = {}
    udctwin[1][1] = _to_sparse(Winlow, param_udct.winthresh)

    param_udct.ind = {}
    param_udct.ind[1] = {}
    param_udct.ind[1][1] = np.zeros((1, 1), dtype=int)
    # every combination of 2 dimension out of 1:dim
    mperms = np.asarray(list(combinations(np.arange(1, param_udct.dim + 1), 2)))
    M = {}
    for ind in range(1, 1 + len(mperms)):
        M[(ind, 1)], M[ind, 2] = adapt_grid(
            Sgrid[mperms[ind - 1, 0]], Sgrid[mperms[ind - 1, 1]]
        )

    # gather angle function for each pyramid

    Mdir = {}
    Mang = {}
    Mang_in = {}
    for res in range(1, param_udct.res + 1):
        Mang[res] = {}
        Mang_in[res] = {}

        Mdir[res] = np.zeros((param_udct.dim, param_udct.dim - 1), dtype=int)
        # for each resolution
        for ind in range(1, param_udct.dim + 1):
            # for each pyramid in resolution res
            cnt = 1
            # cnt is number of angle function required for each pyramid
            # now loop through mperms
            Mdir[res][ind - 1, :] = np.array(
                list(range(1, ind)) + list(range(ind + 1, param_udct.dim + 1)),
                dtype=int,
            )
            # Mdir is dimension of need to calculate angle function on each
            # hyperpyramid
            for hp in range(1, mperms.shape[0] + 1):
                for ndir in range(1, 3):
                    # Fill with zeros, will be replaced
                    # Mang[res][(ind, cnt)] = np.zeros(param_udct.size)
                    if mperms[hp - 1, ndir - 1] == ind:
                        tmp = angle_fun(
                            M[(hp, ndir)],
                            ndir,
                            param_udct.cfg[res - 1, mperms[hp - 1, 3 - ndir - 1] - 1],
                            param_udct.alpha,
                        )
                        Mang[res][(ind, cnt)] = tmp
                        Mang_in[res][(ind, cnt)] = mperms[hp - 1, 1 - 1 : 2]
                        cnt += 1

    # Mang is 1-d angle function for each hyper pyramid (row) and each angle
    # dimension (column)
    for res in range(1, param_udct.res + 1):
        # for each resolution
        subband = {}
        udctwin[res + 1] = {}
        param_udct.ind[res + 1] = {}
        for in1 in range(1, param_udct.dim + 1):
            udctwin[res + 1][in1] = {}
            # for each hyperpyramid
            ang_in = 1
            for in2 in range(1, param_udct.dim - 1 + 1):
                ln = len(Mang[res][(in1, in2)])
                tmp2 = np.arange(ln, dtype=int)[:, None] + 1
                if in2 == 1:
                    ang_in = tmp2
                else:
                    ln2 = ln * len(ang_in)
                    tmp3 = np.kron(ang_in, np.ones((ln, 1), dtype=int))
                    tmp4 = np.kron(np.ones((ang_in.shape[0], 1), dtype=int), tmp2)
                    ang_in = np.c_[tmp3, tmp4]
            lent = ang_in.shape[0]
            ang_inmax = param_udct.cfg[res - 1, Mdir[res][in1 - 1, :] - 1]
            # lent is the smallest number of windows need to calculated on each
            # pyramid
            # ang_inmax is M-1 vector contain number of angle function per each
            # dimension of the hyperpyramid
            ang_ind = 0
            ind = 1
            for in3 in range(1, lent + 1):
                # for each calculated windows function, estimated all the other
                # flipped window functions
                afun = np.ones(param_udct.size, dtype=float)
                afunin = 1
                for in4 in range(1, param_udct.dim - 1 + 1):
                    idx = ang_in.reshape(len(ang_in), -1)[in3 - 1, in4 - 1]
                    tmp = Mang[res][(in1, in4)][idx - 1]
                    # print(f"{tmp.shape}")
                    tmp2 = Mang_in[res][(in1, in4)]
                    afun2 = angle_kron(tmp, tmp2, param_udct)
                    afun *= afun2
                aafun = {}
                ang_in2 = None
                afun = afun * F2d[res + 1]
                afun = np.sqrt(circshift(afun, tuple(s // 4 for s in param_udct.size)))

                # first windows function
                aafun[afunin] = afun

                # index of current angle
                ang_in2 = ang_in[in3 - 1 : in3, :].copy()
                # print(f"{ang_in2.shape=}")

                # all possible flip along different dimension
                for in5 in range(param_udct.dim - 1, 0, -1):
                    lentmp = ang_in2.shape[0]
                    for in6 in range(1, lentmp + 1):
                        if 2 * ang_in2[in6 - 1, in5 - 1] <= ang_inmax[in5 - 1]:
                            ang_in2tmp = ang_in2[in6 - 1 : in6, :].copy()
                            ang_in2tmp[0, in5 - 1] = (
                                ang_inmax[in5 - 1] + 1 - ang_in2[in6 - 1, in5 - 1]
                            )
                            # print(f"{ang_in2.shape=}")
                            # print(f"{ang_in2tmp.shape=}")
                            ang_in2 = np.concatenate((ang_in2, ang_in2tmp), axis=0)
                            # print(f"{ang_in2.shape=}")
                            a = aafun[in6]
                            b = Mdir[res][in1 - 1, in5 - 1]
                            end = max(aafun.keys())
                            aafun[end + 1] = fftflip(a, b - 1)
                #    Mwin[tmp[in3-1,:]] = afun
                # subband[ang]
                aafun = np.concatenate(
                    [aafun[k][None, ...] for k in sorted(aafun.keys())], axis=0
                )
                if isinstance(ang_ind, int) and ang_ind == 0:
                    ang_ind = ang_in2
                    # subband[in1] = aafun.copy()
                    for in7 in range(1, ang_ind.shape[0] + 1):
                        # convert to sparse format
                        udctwin[res + 1][in1][in7] = _to_sparse(
                            aafun[in7 - 1], param_udct.winthresh
                        )
                else:
                    inold = ang_ind.shape[0]
                    ang_ind = np.concatenate((ang_ind, ang_in2), axis=0)
                    innew = ang_ind.shape[0]
                    # note:
                    # subband has to be expanded to accommodate new indices
                    # matlab does that automatically, but numpy does not
                    # the following line DOES NOT WORK
                    # subband[in1][inold + 1 - 1 : innew] = aafun
                    for in7 in range(inold + 1, innew + 1):
                        in8 = in7 - inold
                        udctwin[res + 1][in1][in7] = _to_sparse(
                            aafun[in8 - 1], param_udct.winthresh
                        )
                    param_udct.ind[res + 1][in1] = ang_ind.copy()

    sumw2 = np.zeros(param_udct.size)
    idx = udctwin[1][1][:, 0].astype(int) - 1
    val = udctwin[1][1][:, 1]
    sumw2.T.flat[idx] += val.T.ravel() ** 2
    for res in range(1, param_udct.res + 1):
        for dir in range(1, param_udct.dim + 1):
            for ang in range(1, len(udctwin[res + 1][dir]) + 1):
                tmpw = np.zeros(param_udct.size)
                idx = udctwin[res + 1][dir][ang][:, 0].astype(int) - 1
                val = udctwin[res + 1][dir][ang][:, 1]
                tmpw.T.flat[idx] += val.T.ravel() ** 2
                sumw2 += tmpw
                tmpw = fftflip(tmpw, dir - 1)
                sumw2 += tmpw

    sumw2 = np.sqrt(sumw2)
    idx = udctwin[1][1][:, 0].astype(int) - 1
    udctwin[1][1][:, 1] /= sumw2.T.ravel()[idx]
    for res in range(1, param_udct.res + 1):
        for dir in range(1, param_udct.dim + 1):
            for ang in range(1, len(udctwin[res + 1][dir]) + 1):
                idx = udctwin[res + 1][dir][ang][:, 0].astype(int) - 1
                val = udctwin[res + 1][dir][ang][:, 1]
                udctwin[res + 1][dir][ang][:, 1] /= sumw2.T.ravel()[idx]

    # decimation ratio for each band
    param_udct.dec = {}
    for res in range(1, param_udct.res + 1):
        tmp = np.ones((param_udct.dim, param_udct.dim))
        param_udct.dec[res] = 2.0 ** (param_udct.res - res + 1) * tmp
        for ind in range(1, param_udct.dim + 1):
            ind2 = Mdir[res][ind - 1, :] - 1
            ind3 = Mdir[res][ind - 1, :] - 1
            param_udct.dec[res][ind - 1, ind2] = (
                2.0 ** (param_udct.res - res) * 2 * param_udct.cfg[res - 1, ind3] / 3
            )

    param_udct.len_r = param_udct.len / 2.0 ** ((param_udct.res - 1) * param_udct.dim)
    for res in range(1, param_udct.res + 1):
        for ind in range(1, param_udct.dim + 1):
            a = udctwin[res + 1][ind]
            b = param_udct.dec[res][ind - 1, :]
            param_udct.len_r += len(a) * 2 * param_udct.len / np.prod(b)

    # sort the window
    newwin = {}
    for res in range(2, param_udct.res + 1 + 1):
        for pyr in range(1, param_udct.dim + 1):
            # take out the angle index list
            mlist = param_udct.ind[res][pyr].copy()

            # map it to a number
            mult = 1
            nlist = np.zeros((mlist.shape[0], 1))
            for d in range(mlist.shape[1], 0, -1):
                for b in range(1, mlist.shape[0] + 1):
                    nlist[b - 1] += mult * mlist[b - 1, d - 1]
                mult *= 100
            ix = np.argsort(nlist, axis=0) + 1
            # b = nlist[ix]

            newind = mlist.copy()
            for b in range(1, mlist.shape[0] + 1):
                newind[b - 1, :] = mlist[ix[b - 1] - 1, :].copy()
                newwin[b] = udctwin[res][pyr][ix[b - 1].item()].copy()

            param_udct.ind[res][pyr] = newind.copy()
            udctwin[res][pyr] = newwin.copy()

    return udctwin
